/* -*- mode: c; c-basic-offset: 4; -*- */

#if HAVE_CONFIG_H
#include "daq_config.h"
#endif /* HAVE_CONFIG_H */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#else
#include <io.h>
#endif /* HAVE_UNISTD_H */

#include "daqc.h"
#include "daqc_response.h"
#include "daqc_internal.h"
#include "daqc_net.h"
#include "nds_auth.h"
#include "nds_os.h"
#include "nds2.h"

/*
 *  Connect to the DAQD server on the host identified by `ip' address.
 *  Returns zero if OK or the error code if failed.
 */
int
nds2_connect (daq_t *daq, const char* host, int port) 
{
    int on = 1, resp;
    int connect_rc = 0;
    int cntr;

    /*  Get a socket.
     */
    daq->sockfd = socket (AF_INET, SOCK_STREAM, 0);
    if (daq->sockfd < 0) {
	perror("socket() error");
	return DAQD_SOCKET;
    }

    setsockopt (daq->sockfd, SOL_SOCKET, SO_REUSEADDR, 
		(const void *) &on, (socklen_t) sizeof (on));

    /*----------------------------------  Get the server address.     */
    resp = daq_set_server(daq, host, port);
    if (resp) {
	perror("Error in daq_set_server");
	close(daq->sockfd);
	daq->sockfd = -1;
	return resp;
    }

    /*-----------------------------------  Connect to the server      */
#if _WIN32
#if 0
    {
	long flag = 1;
	
	ioctlsocket( daq->sockfd, FIONBIO, &flag );
    }
#endif /* 0 */
#else
    fcntl(daq -> sockfd, F_SETFL, O_NONBLOCK);
#endif /* _WIN32 */
    printf("Connecting NDS2 ..");
    fflush(stdout);
#define MAX_CONNECT_RETRY 4
    connect_rc = 0;
    for (cntr=0; cntr<MAX_CONNECT_RETRY; ++cntr) {
	connect_rc = connect (daq->sockfd, 
			      (struct sockaddr *) daq -> srvr_addr, 
			      (socklen_t) sizeof (*(daq -> srvr_addr))
			      );
	if (connect_rc < 0 && (errno == EINPROGRESS || errno == EALREADY)) {
	    sleep(1);
	    printf("..");
	    fflush(stdout);
	} else {
	    break;
	}
    }
    if (connect_rc < 0 && errno != EISCONN) {
	printf(" failed\n");
	perror("connect() error");
#if ! _WIN32
	close (daq -> sockfd);
#endif /* ! _WIN32 */
	daq->sockfd = -1;
	return DAQD_CONNECT;
    }

    printf(" authenticate");    
#if _WIN32
#if 0
    {
	long flag = 0;
	
	ioctlsocket( daq->sockfd, FIONBIO, &flag );
    }
#endif /* 0 */
#else /* _WIN32 */
    fcntl(daq -> sockfd, F_SETFL, 0);
#endif /* _WIN32 */
    resp = daq_send(daq, "authorize");
    if (resp == DAQD_SASL) {
	printf(" ...");
	resp = nds_authenticate(daq, host);
	if (resp) resp = DAQD_SASL;
    }

    if (resp == DAQD_OK) {
	printf(" done\n");
	daq->datafd = daq->sockfd;
    } else {
	printf(" failed: %s\n", daq_strerror(resp));
	close (daq->sockfd);
	daq->sockfd = -1;
    }
    return resp;
}


/*
 *  Disconnect from the server and close the socket file descriptor
 */
int
nds2_disconnect (daq_t *daq) {
    /*  Send a quit command */
    char *command = "quit;\n";
    size_t   lcmd = strlen(command);
    if (write (daq -> sockfd, command, lcmd) != lcmd) {
#if DAQC_ACCESS_VERBOSE_ERRORS
	fprintf (stderr, "daq_disconnect: write errno=%d\n", errno);
#endif
	return DAQD_WRITE;
    }
    usleep(100000);
    nds_auth_disconnect(daq);
    close (daq -> sockfd);
    return 0;
}

/*
 *  Receive channel data using the NDS2 protocol
 */
int
nds2_recv_channel_list(daq_t *daq, daq_channel_t *channel, int num_channels, 
		   int *num_channels_received, time_t gps, enum chantype type)
{
#define BUF_LENGTH 512
#define MAX_TYPESTR_LENGTH 16
    char buf [BUF_LENGTH];
    char typestr [MAX_TYPESTR_LENGTH];
    int resp, fd, i;
    int4_type	channels;

    /*----------------------------------  Issue the request               */
    if (num_channels != 0) {
#if HAVE_SPRINTF_S
#if DAQC_ACCESS_VERBOSE_ERRORS
	fprintf (stderr, "nds2_recv_channel_list: sizeof( buf ):%d %s - %d\n",
		 sizeof( buf ),
		 __FILE__,
		 __LINE__ );
#endif
	sprintf_s(buf, sizeof( buf ),
		  "get-channels %i %s;", (int)gps, cvt_chantype_str(type));
#else	/* HAVE_SPRINTF_S */
	sprintf(buf,
		"get-channels %i %s;", (int)gps, cvt_chantype_str(type));
#endif /* HAVE_SPRINTF_S */
#if DAQC_ACCESS_VERBOSE_ERRORS
	printf("DEBUG: Requesting channel list\n");
#endif /* DAQC_ACCESS_VERBOSE_ERRORS */
    } else {
#if HAVE_SPRINTF_S
#if DAQC_ACCESS_VERBOSE_ERRORS
	fprintf (stderr, "nds2_recv_channel_list: sizeof( buf ):%d %s - %d\n",
		 sizeof( buf ),
		 __FILE__,
		 __LINE__ );
#endif
	sprintf_s(buf, sizeof( buf ),
		  "count-channels %i %s;",(int)gps, cvt_chantype_str(type));
#else /* HAVE_SPRINTF_S */
	sprintf(buf,
		"count-channels %i %s;",(int)gps, cvt_chantype_str(type));
#endif /* HAVE_SPRINTF_S */
#if DAQC_ACCESS_VERBOSE_ERRORS
	printf("DEBUG: Requesting channel count\n");
#endif /* DAQC_ACCESS_VERBOSE_ERRORS */
    }
#if DAQC_ACCESS_VERBOSE_ERRORS
    printf("DEBUG: buf - %s\n", buf );
#endif /* DAQC_ACCESS_VERBOSE_ERRORS */
    resp = daq_send(daq, buf);
    if (resp) {
	printf("Error return from daq_send(): %s\n", daq_strerror(resp));
	return resp;
    }

    /*----------------------------------  Get the number of channels      */
    fd = daq->sockfd;
    if (read_uint4(fd, (uint4_type*)&channels) || channels < 0) {
	fprintf (stderr, "Couldn't determine the number of data channels\n");
	return DAQD_ERROR;
    }
    *num_channels_received = channels;
#if DAQC_ACCESS_VERBOSE_ERRORS
    printf("DEBUG: channels: %d num_channels_received: %d\n", 
	   channels, *num_channels_received );
#endif /* DAQC_ACCESS_VERBOSE_ERRORS */
    if (num_channels == 0) return DAQD_OK;

    /*----------------------------------  Loop over channels              */
    for (i=0; i < channels; i++) {
	enum chantype ctype;
	double rate;
	int len=_daq_read_string(fd, BUF_LENGTH, buf);
	daq_data_t dtype;

#if DAQC_ACCESS_VERBOSE_ERRORS
	printf( "DEBUG: i: %d len: %d\n", i, len );
#endif /* DAQC_ACCESS_VERBOSE_ERRORS */
	if (len <= 0) return DAQD_ERROR;
	if (i < num_channels) {
	    const char* p = buf;
	    const char* end = buf + len;
	    char name[MAX_CHANNEL_NAME_LENGTH];
	    _daq_get_string(&p, end, name, sizeof(name));
	    _daq_get_string(&p, end, typestr, sizeof(typestr));
	    ctype = cvt_str_chantype(typestr);
	    rate = strtod(p, (char**)&p);
	    _daq_get_string(&p, end, typestr, sizeof(typestr));
	    dtype = data_type_code(typestr);
	    daq_init_channel(channel + i, name, ctype, rate, dtype);
	}
    }
#if DAQC_ACCESS_VERBOSE_ERRORS
    printf("DEBUG: Channel List worked OK\n");
#endif /* DAQC_ACCESS_VERBOSE_ERRORS */
    return DAQD_OK;
}

/*    Get the name string length
 */
static size_t
_nds2_name_list_length(daq_t *daq) {
    uint4_type i;
    uint4_type N = daq->num_chan_request;
    size_t namel = 0;
    chan_req_t* req = daq->chan_req_list;
    for (i=0; i<N; ++i) {
	chantype_t ctype;

	namel += strlen(req->name) + 1;
	ctype = req->type;
	if (ctype != cUnknown) namel += strlen(cvt_chantype_str(ctype)) + 1;
	req++;
    }
    return namel;
}

/*    Get the name string length
 */
static size_t
_nds2_get_name_list(daq_t *daq, char* p) {
    chan_req_t* req;

    size_t i = 0;
    size_t j = 0;
    size_t N = daq->num_chan_request;
    p[i++] = '{';
    req = daq->chan_req_list;
    for (j=0; j<N; j++) {
	const char* pchn = req->name;
	chantype_t ctype;

	if (j) p[i++] = ' ';
	while (*pchn) p[i++] = *pchn++;
	ctype = req->type;
	if (ctype != cUnknown) {
	    p[i++] = ',';
	    pchn = cvt_chantype_str(ctype);
	    while (*pchn) p[i++] = *pchn++;
	}
	req++;
    }
    p[i++] = '}';
    return i;
}

/*
 *  Receive source list using the NDS2 protocol
 */
int 
nds2_recv_source_list(daq_t *daq, char* list, size_t max_len, time_t gps, 
		      long* str_len) {
    int    rc = 0;
    size_t namel, txt_len;
    char   *cmd, *cmd_txt, *p; 

    *str_len = 0;
    namel = _nds2_name_list_length(daq);

    /*----------------------------------  Build up a command
     */
    cmd = "get-source-list";

    /*----------------------------------  Get the text length, allocate a buffer
     */
    txt_len = strlen(cmd)
	           + strlen(" xxxxxxxxxx {};") 
	           + namel + daq->num_chan_request;
    cmd_txt = (char*)malloc(txt_len);

    /*----------------------------------  Build the command
     */
    p = cmd_txt;
    while (*cmd) *p++ = *cmd++;
    *p++ = ' ';
    p += daq_cvt_timet_string(gps, p);
    *p++ = ' ';
    p += _nds2_get_name_list(daq, p);
    *p++ = ';';
    *p++ = 0;  

    /*----------------------------------  Send the command free the command
     */
    rc = daq_send(daq, cmd_txt);
    free(cmd_txt);
    cmd_txt = 0;

    if (rc) {
	printf("Error in daq_send: %s\n", daq_strerror(rc));
	return rc;
    }

    /*----------------------------------  Read the source string      */
    *str_len = _daq_read_string(daq->sockfd, max_len, list);
    if (*str_len < 0) return DAQD_ERROR;
    return rc;
}

/*  nds2_request_data(daq_t*, time_t, time_t)
 *
 *  Get requested channel data for the specified interval.
 */
int
nds2_request_data(daq_t* daq, time_t start, time_t end, time_t dt) {
    int    wid, bl, rc = DAQD_OK;
    size_t N, i, namel, txt_len;
    char   *cmd, *cmd_txt, *p;
    chan_req_t* req;

    /*  Get the request type and sum channel name length
     */
    N = daq->num_chan_request;
    if (N == 0) {
#if DAQC_ACCESS_VERBOSE_ERRORS
	fprintf(stderr, "No channels requested\n");
#endif
	return DAQD_ERROR;
    }

    namel = _nds2_name_list_length(daq);

    /*----------------------------------  Build up a command
     */
    cmd = "get-data";

    /*----------------------------------  Get the text length, allocat a buffer
     */
    txt_len = strlen(cmd)
	    + strlen(" xxxxxxxxxx xxxxxxxxx xxxxxxxxxx {};") 
	    + namel + N;
    cmd_txt = (char*)malloc(txt_len);

    /*----------------------------------  Build the command
     */
    p = cmd_txt;
    while (*cmd) *p++ = *cmd++;
    *p++ = ' ';
    p += daq_cvt_timet_string(start, p);
    *p++ = ' ';
    p += daq_cvt_timet_string(end, p);
    *p++ = ' ';
    p += daq_cvt_timet_string(dt, p);
    *p++ = ' ';
    p += _nds2_get_name_list(daq, p);
    *p++ = ';';
    *p++ = 0;  

    /*----------------------------------  Reset status words before request
     */
    req = daq->chan_req_list;
    for (i=0; i<N; ++i) {
	req[i].status = 0;
    }

    /*----------------------------------  Send the command free the command
     */
    rc = daq_send(daq, cmd_txt);
    free(cmd_txt);
    cmd_txt = 0;


    if (rc) {
	printf("Error in daq_send: %s\n", daq_strerror(rc));
	return rc;
    }

    /*------------------------------  Get the writer ID
     */
    wid = daq_recv_id(daq);
    if (wid < 0) {
	printf("Error reading writerID in daq_recv_block\n");
	return DAQD_ERROR;
    }

    /*------------------------------  Get the offline flag
     */
    bl = daq_recv_block_num(daq);
    if (bl < 0) {
	printf("Error reading offline flag in daq_recv_block\n");
	rc = DAQD_ERROR;
    }
    return rc;
}


struct nds2_chan_status {
    int status;
    int offset;
    int data_type;
    float rate;
    float signal_offset;
    float signal_slope;
};

/*  nds2_receive_reconfigure(daq_t* daq, long block_len)
 *
 *  Receive a reconfigure block. receive_reconfigure is received after the 
 *  block header has been read in. The block length does not include the
 *  header length.
 */
int 
nds2_receive_reconfigure(daq_t* daq, size_t block_len) {
    int fd;
    size_t nchannels, i;
    chan_req_t* list;

    nchannels = block_len / sizeof(struct nds2_chan_status);
    if (nchannels * sizeof(struct nds2_chan_status) != block_len) {
	fprintf (stderr, 
		 "Channel reconfigure block length has bad length (%zi)\n",
		 block_len);
    }
    if (nchannels != daq->num_chan_request) return -1;

    fd = daq->datafd;
    list = daq->chan_req_list;
    for (i = 0; i < nchannels; i++) {
	float rate;
	uint4_type dtype;
	if (read_uint4(fd, (uint4_type*)&list[i].status))  return DAQD_ERROR;
	if (read_uint4(fd, &list[i].offset))  return DAQD_ERROR;
	if (read_uint4(fd, &dtype))           return DAQD_ERROR;
	if (list[i].type == cUnknown) list[i].type = dtype >> 16;
	list[i].data_type = (dtype & 0xffff);
	if (read_float(fd, &rate))           return DAQD_ERROR;
	list[i].rate = rate;
	if (read_float(fd, &list[i].s.signal_offset))  return DAQD_ERROR;
	if (read_float(fd, &list[i].s.signal_slope))   return DAQD_ERROR;
    }
    return -2; 
}

/*
 *  User authentication
 */
int
nds2_startup(void) {
    return nds_auth_startup();
}
